/*******************************************************************************
 * --------------------------------------------------------------------------- *
 * File: * @(#) SAX2ScanHandler.java * Author: * Robert M. Hubley
 * rhubley@systemsbiology.org
 * ****************************************************************************** * * *
 * This software is provided ``AS IS'' and any express or implied * *
 * warranties, including, but not limited to, the implied warranties of * *
 * merchantability and fitness for a particular purpose, are disclaimed. * * In
 * no event shall the authors or the Institute for Systems Biology * * liable
 * for any direct, indirect, incidental, special, exemplary, or * *
 * consequential damages (including, but not limited to, procurement of * *
 * substitute goods or services; loss of use, data, or profits; or * * business
 * interruption) however caused and on any theory of liability, * * whether in
 * contract, strict liability, or tort (including negligence * * or otherwise)
 * arising in any way out of the use of this software, even * * if advised of
 * the possibility of such damage. * * *
 * ******************************************************************************
 * 
 * ChangeLog
 * 
 * $Log: SAX2ScanHandler.java,v $ Revision 1.1.1.1 2003/04/09 00:02:54 ppatrick
 * Initial import.
 * 
 * 10-05-2004: fixed bug in for loop, M. Vogelzang
 * 
 * 1.2 added logging & handling of precursorMZ 
 * M. Vogelzang  
 * (logging commented out to prevent need of log4j library)
 * 
 ******************************************************************************/
package org.systemsbiology.jrap;
//import org.apache.log4j.Logger;
import java.util.zip.DataFormatException;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

public class SAX2ScanHandler extends DefaultHandler
{
//	private static final Logger logger =
//		Logger.getLogger(SAX2ScanHandler.class);
	/** A new Scan Object */
	protected Scan tmpScan;

	/** A string to hold the Base64 peak data */
	protected StringBuffer peakData = new StringBuffer();

	/** Flag to indicate we are reading a peak tag */
	protected boolean inPeak = false;

	/** Buffer to hold characters while getting precursorMZ value */
	protected StringBuffer precursorBuffer;

	/** Flag to indicate if we are reading the precursor MZ value */
	protected boolean inPrecursorMZ = false;
	
	protected boolean isZLibCompressed = false;
	
	protected int compressedLen = -1;

	public SAX2ScanHandler() {
		
	}
	
	//
	// Getters
	//

	public Scan getScan()
	{
		return (tmpScan);
	}

	protected String getStringAttribute(Attributes attrs, String name)
	{
		return attrs.getValue(name);
	}
	
	protected int getIntAttribute(Attributes attrs, String name)
	{
		int result;

		if (attrs.getValue(name) == null) // attribute not present
			return -1;

		try
		{
			result = Integer.parseInt(attrs.getValue(name));
		} catch (NumberFormatException e)
		{
//			logger.error("Numberformatexception!", e);
			result = -1;
		}
		return (result);
	}

	protected float getFloatAttribute(Attributes attrs, String name)
	{
		float result;

		if (attrs.getValue(name) == null) // attribute not present
			return -1;

		try
		{
			result = Float.parseFloat(attrs.getValue(name));
		} catch (NumberFormatException e)
		{
//			logger.error("Numberformatexception!", e);
			result = -1;
		} catch (NullPointerException e1)
		{
//			logger.error("Nullpointerexception!", e1);
			result = -1;
		}
		return (result);
	}

	//
	// ContentHandler Methods
	//

	/** Start document. */
	public void startDocument() throws SAXException
	{
		// Nothing to do
	} // startDocument()

	protected Scan newScan() {
		return new Scan();
	}
	
	private void addScanAttributes(Attributes attrs) {

		tmpScan.setNum(getIntAttribute(attrs, "num"));
		tmpScan.setMsLevel(getIntAttribute(attrs, "msLevel"));
		tmpScan.setPeaksCount(getIntAttribute(attrs, "peaksCount"));
		tmpScan.setPolarity(attrs.getValue("polarity"));
		tmpScan.setScanType(attrs.getValue("scanType"));
		tmpScan.setCentroided(getIntAttribute(attrs, "centroided"));
		tmpScan.setDeisotoped(getIntAttribute(attrs, "deisotoped"));
		tmpScan.setChargeDeconvoluted(
			getIntAttribute(attrs, "chargeDeconvoluted"));
		tmpScan.setRetentionTime(attrs.getValue("retentionTime"));
		tmpScan.setStartMz(getFloatAttribute(attrs, "startMz"));
		tmpScan.setEndMz(getFloatAttribute(attrs, "endMz"));
		tmpScan.setLowMz(getFloatAttribute(attrs, "lowMz"));
		tmpScan.setHighMz(getFloatAttribute(attrs, "highMz"));
		tmpScan.setBasePeakMz(getFloatAttribute(attrs, "basePeakMz"));
		tmpScan.setBasePeakIntensity(
			getFloatAttribute(attrs, "basePeakIntensity"));
		tmpScan.setTotIonCurrent(getFloatAttribute(attrs, "totIonCurrent"));
	}
	
	protected void addScanPrecursorAttributes(Attributes attrs) {
		tmpScan.setPrecursorScanNum(
				getIntAttribute(attrs, "precursorScanNum"));
		tmpScan.setPrecursorCharge(
				getIntAttribute(attrs, "precursorCharge"));
		tmpScan.setCollisionEnergy(
				getFloatAttribute(attrs, "collisionEnergy"));
		tmpScan.setIonisationEnergy(
				getFloatAttribute(attrs, "ionisationEnergy"));
	}
	
	/** Start element. */
	public void startElement(
		String uri,
		String local,
		String raw,
		Attributes attrs)
		throws SAXException
	{
		if (raw.equals("scan"))
		{
			tmpScan = newScan();
			addScanAttributes(attrs);
		} else if (raw.equals("peaks"))
		{
			/* 
			 * The peaks data (the actual spectrum data points represented using one of the above
			 * contentTypes) may optionally be compressed during conversion using the lossless 
			 * data-compression library zlib.  Zlib is a free library available for a broad range
			 * of disparate computer hardware and operating systems. 
			 * mass spectrometry data that is approx. 20% smaller in size than the original 
			 * RAW file. 
			 */

			String byteOrder = getStringAttribute(attrs, "byteOrder") ;
			String compressionType = getStringAttribute(attrs, "compressionType") ;
			compressedLen = getIntAttribute(attrs, "compressedLen");
			
			if (tmpScan.getPeaksCount() > 0 && (byteOrder == null || !byteOrder.equals("network"))) {
				throw new IllegalArgumentException("the 'byteOrder' value parameter " +
						" is not valid, unknown little/big endian type");
			}
			
			if (compressionType != null && !compressionType.equals("none")) {
				if (!compressionType.equals("zlib")) {
					throw new IllegalArgumentException("the 'compressionType' value parameter " +
							compressionType +" is not valid");
				}
				if (compressedLen == -1) {
					throw new IllegalArgumentException("the 'compressedLen' value parameter " +
							" is not defined");
				}
				isZLibCompressed = true;
			}
			int precision = getIntAttribute(attrs, "precision");
			
			if ((precision & 32) == 0 && (precision & 64) == 0) {
				throw new IllegalArgumentException("the 'precision' value parameter " +
						" is not well defined");
			}
			
			tmpScan.setPrecision(precision);
			inPeak = true;
		} else if (raw.equals("precursorMz"))
		{
			addScanPrecursorAttributes(attrs);
			
			precursorBuffer = new StringBuffer();
			inPrecursorMZ = true;
		}
	} // startElement(String,String,StringAttributes)
	
	public void endElement(String uri, String local, String raw)
		throws SAXException
	{
		if (raw.equals("peaks"))
		{				
			
			if (tmpScan.getPeaksCount() > 0) {

				// if zlib compressed spectrum then:
				//   1. base64 decoding (from base64 char to bytes)
				byte[] bytes = Base64.decode(peakData.toString());

				//   2. zlib decompression			
				if (isZLibCompressed) {
					// 2 values per peak, precision/8 bytes per value
					int uncompressedLen = (tmpScan.getPeaksCount() * tmpScan.getPrecision()/4);

					try {
						bytes = Bytes.uncompress(bytes, compressedLen, uncompressedLen);
					} catch (DataFormatException e) {
						throw new SAXException(e.getMessage());
					}
				}

				if (tmpScan.getPeaksCount() > 0) {
					float[][] tmpMassIntensityList =
						Bytes.byteQuadsToMzInts(bytes, 
								tmpScan.getPeaksCount(), tmpScan.getPrecision() / 8);

					tmpScan.setMassIntensityList(tmpMassIntensityList);
					tmpMassIntensityList = null;
				}
				
			}
			inPeak = false;
			peakData.delete(0, peakData.capacity());
			throw (new SAXException("ScanEndFoundException"));			
		} else if (raw.equals("precursorMz"))
		{
			
			if (precursorBuffer.length() == 0) {
				System.out.println("prec = "+precursorBuffer);
				System.out.println("empty string: "+tmpScan.getNum());
			}
			
			tmpScan.setPrecursorMz(
				Float.parseFloat(precursorBuffer.toString()));
			precursorBuffer = null; // make available for garbage collection

			inPrecursorMZ = false;
		}
	} // endElement()

	/** Characters. */
	public void characters(char ch[], int start, int length)
		throws SAXException
	{
		if (inPeak)
		{
			peakData.append(ch, start, length);
		} else if (inPrecursorMZ)
		{
			precursorBuffer.append(ch, start, length);
		}
	} // characters(char[],int,int);

	/** Ignorable whitespace. */
	public void ignorableWhitespace(char ch[], int start, int length)
		throws SAXException
	{
		// Do nothing
	} // ignorableWhitespace(char[],int,int);

	/** Processing instruction. */
	public void processingInstruction(String target, String data)
		throws SAXException
	{
		// Do nothing
	} // processingInstruction(String,String)

	//
	// ErrorHandler methods
	//

	/** Warning. */
	public void warning(SAXParseException ex) throws SAXException
	{
		// Do nothing
		//printError("Warning", ex);
	} // warning(SAXParseException)

	/** Error. */
	public void error(SAXParseException ex) throws SAXException
	{
		// Do nothing
		//printError("Error", ex);
	} // error(SAXParseException)

	/** Fatal error. */
	public void fatalError(SAXParseException ex) throws SAXException
	{
		// Do nothing
		//printError("Fatal Error", ex);
	} // fatalError(SAXParseException)

	//
	// Protected methods
	//

	/** Prints the error message. */
	protected void printError(String type, SAXParseException ex)
	{
		System.err.print("[");
		System.err.print(type);
		System.err.print("] ");
		if (ex == null)
		{
			System.out.println("!!!");
		}
		String systemId = ex.getSystemId();
		if (systemId != null)
		{
			int index = systemId.lastIndexOf('/');
			if (index != -1)
				systemId = systemId.substring(index + 1);
			System.err.print(systemId);
		}
		System.err.print(':');
		System.err.print(ex.getLineNumber());
		System.err.print(':');
		System.err.print(ex.getColumnNumber());
		System.err.print(": ");
		System.err.print(ex.getMessage());
		System.err.println();
		System.err.flush();
	} // printError(String,SAXParseException)

}
