package org.systemsbiology.jrap.extension;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

public class ExtSAX2ScanHeaderHandler extends DefaultHandler
{
	//		private static final Logger logger =
	//			Logger.getLogger(SAX2ScanHeaderHandler.class);

	/** A new ScanHeader Object */
	protected ExtScanHeader tmpScanHeader;

	/** Buffer to hold characters while getting precursorMZ value */
	protected StringBuffer precursorBuffer;

	/** Flag to indicate if we are reading the precursor MZ value */
	protected boolean inPrecursorMZ = false;

	//
	// Getters
	//

	public ExtScanHeader getScanHeader()
	{
		return (tmpScanHeader);
	}

	private String getStringAttribute(Attributes attrs, String name)
	{
		return attrs.getValue(name);
	}
	
	private int getIntAttribute(Attributes attrs, String name)
	{
		int result;

		if (attrs.getValue(name) == null) // attribute not present
			return -1;

		try
		{
			result = Integer.parseInt(attrs.getValue(name));
		} catch (NumberFormatException e)
		{
			//				logger.error("Numberformatexception!", e);
			result = -1;
		}
		return (result);
	}

	private float getFloatAttribute(Attributes attrs, String name)
	{
		float result;

		if (attrs.getValue(name) == null) // attribute not present
			return -1;

		try
		{
			result = Float.parseFloat(attrs.getValue(name));
		} catch (NumberFormatException e)
		{
			//				logger.error("Numberformatexception!", e);
			result = -1;
		} catch (NullPointerException e1)
		{
			//				logger.error("Nullpointerexception!", e1);
			result = -1;
		}
		return (result);
	}

	//
	// ContentHandler Methods
	//

	/** Start document. */
	public void startDocument() throws SAXException
	{
		// Nothing to do
	} // startDocument()

	/** Start element. */
	public void startElement(
			String uri,
			String local,
			String raw,
			Attributes attrs)
	throws SAXException
	{
		if (raw.equals("scan"))
		{
			tmpScanHeader = new ExtScanHeader();
			tmpScanHeader.setNum(getIntAttribute(attrs, "num"));
			tmpScanHeader.setMsLevel(getIntAttribute(attrs, "msLevel"));
			tmpScanHeader.setPeaksCount(getIntAttribute(attrs, "peaksCount"));
			tmpScanHeader.setPolarity(attrs.getValue("polarity"));
			tmpScanHeader.setScanType(attrs.getValue("scanType"));
			tmpScanHeader.setCentroided(getIntAttribute(attrs, "centroided"));
			tmpScanHeader.setDeisotoped(getIntAttribute(attrs, "deisotoped"));
			tmpScanHeader.setChargeDeconvoluted(
					getIntAttribute(attrs, "chargeDeconvoluted"));
			tmpScanHeader.setRetentionTime(attrs.getValue("retentionTime"));
			tmpScanHeader.setStartMz(getFloatAttribute(attrs, "startMz"));
			tmpScanHeader.setEndMz(getFloatAttribute(attrs, "endMz"));
			tmpScanHeader.setLowMz(getFloatAttribute(attrs, "lowMz"));
			tmpScanHeader.setHighMz(getFloatAttribute(attrs, "highMz"));
			tmpScanHeader.setBasePeakMz(getFloatAttribute(attrs, "basePeakMz"));
			tmpScanHeader.setBasePeakIntensity(
					getFloatAttribute(attrs, "basePeakIntensity"));
			tmpScanHeader.setTotIonCurrent(getFloatAttribute(attrs, "totIonCurrent"));
		} else if (raw.equals("peaks"))
		{
			tmpScanHeader.setPrecision(getIntAttribute(attrs, "precision"));
			throw (new SAXException("ScanHeaderEndFoundException"));
		} else if (raw.equals("precursorMz"))
		{
			tmpScanHeader.setPrecursorScanNum(
					getIntAttribute(attrs, "precursorScanNum"));
			tmpScanHeader.setPrecursorCharge(
					getIntAttribute(attrs, "precursorCharge"));
			tmpScanHeader.setActivationMethod(
					getStringAttribute(attrs, "activationMethod"));
			tmpScanHeader.setPrecursorIntensity(
					getFloatAttribute(attrs, "precursorIntensity"));
			tmpScanHeader.setCollisionEnergy(
					getFloatAttribute(attrs, "collisionEnergy"));
			tmpScanHeader.setIonisationEnergy(
					getFloatAttribute(attrs, "ionisationEnergy"));

			precursorBuffer = new StringBuffer();
			inPrecursorMZ = true;
		}
	} // startElement(String,String,StringAttributes)

	public void endElement(String uri, String local, String raw)
	throws SAXException
	{
		if (raw.equals("precursorMz"))
		{
			tmpScanHeader.setPrecursorMz(
					Float.parseFloat(precursorBuffer.toString()));
			precursorBuffer = null; // make available for garbage collection

			inPrecursorMZ = false;
		}
	} // endElement()

	/** Characters. */
	public void characters(char ch[], int start, int length)
	throws SAXException
	{
		if (inPrecursorMZ)
		{
			precursorBuffer.append(ch, start, length);
		}
	} // characters(char[],int,int);

	/** Ignorable whitespace. */
	public void ignorableWhitespace(char ch[], int start, int length)
	throws SAXException
	{
		// Do nothing
	} // ignorableWhitespace(char[],int,int);

	/** Processing instruction. */
	public void processingInstruction(String target, String data)
	throws SAXException
	{
		// Do nothing
	} // processingInstruction(String,String)

	//
	// ErrorHandler methods
	//

	/** Warning. */
	public void warning(SAXParseException ex) throws SAXException
	{
		// Do nothing
		//printError("Warning", ex);
	} // warning(SAXParseException)

	/** Error. */
	public void error(SAXParseException ex) throws SAXException
	{
		// Do nothing
		//printError("Error", ex);
	} // error(SAXParseException)

	/** Fatal error. */
	public void fatalError(SAXParseException ex) throws SAXException
	{
		// Do nothing
		//printError("Fatal Error", ex);
	} // fatalError(SAXParseException)

	//
	// Protected methods
	//

	/** Prints the error message. */
	protected void printError(String type, SAXParseException ex)
	{
		System.err.print("[");
		System.err.print(type);
		System.err.print("] ");
		if (ex == null)
		{
			System.out.println("!!!");
		}
		String systemId = ex.getSystemId();
		if (systemId != null)
		{
			int index = systemId.lastIndexOf('/');
			if (index != -1)
				systemId = systemId.substring(index + 1);
			System.err.print(systemId);
		}
		System.err.print(':');
		System.err.print(ex.getLineNumber());
		System.err.print(':');
		System.err.print(ex.getColumnNumber());
		System.err.print(": ");
		System.err.print(ex.getMessage());
		System.err.println();
		System.err.flush();
	} // printError(String,SAXParseException)

}